/* 
 * $Id: jpeg2v4l.c,v 1.8 2004/02/21 03:53:21 harmen Exp $
 *
 * pccam880/app/jpeg2v4l
 *
 * Copyright (C) 2004 Harmen van der Wal <harmen@harmwal.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <setjmp.h>
#include <errno.h>

#include <jpeglib.h>
#include <linux/videodev.h>

#define ERR(fmt,args...) fprintf(stderr, "%s %s (%d): "fmt, __FILE__, __FUNCTION__, __LINE__,##args)


char *file_in = "/dev/stdin", *file_out;
int fd_out;

void (*orig_output_message)(j_common_ptr cinfo);

int v4l(int width, int height)
{
  struct video_window vwin;
  struct video_picture vpic;

  fd_out = open(file_out, O_WRONLY);
  if (fd_out < 1) {
    ERR("open %s: %s\n", file_out, strerror(errno));
    return -1;
  }

  vpic.colour = 0xffff;
  vpic.hue = 0xffff;
  vpic.brightness = 0xffff;
  vpic.contrast = 0xffff;
  vpic.whiteness = 0xffff;
  vpic.depth = 3;
  vpic.palette = VIDEO_PALETTE_RGB24;

  vwin.x = 0;
  vwin.y = 0;
  vwin.width = width;
  vwin.height = height;
  vwin.chromakey = 0;
  vwin.flags = 0;
  vwin.clipcount = 0;
  
  if (ioctl(fd_out, VIDIOCSPICT, &vpic) < 0) {
    ERR("ioctl %s: %s\n", file_out, strerror(errno));
    return -1;
  }

  if (ioctl (fd_out, VIDIOCSWIN, &vwin) < 0) {
    ERR("ioctl %s: %s\n", file_out, strerror(errno));
    return -1;
  }

  return 0;
}

void rgb(char *buf, int size)
{
  int i;
  char c;
  for(i = 0; i < size - 3; i += 3) {
    c = buf[i];
    buf[i] = buf[i + 2];
    buf[i + 2] = c;
  }
}

void libjpeg_output_message(j_common_ptr cinfo)
{
  ERR("");
  orig_output_message(cinfo);
}

int main(int argc, char **argv)
{
  FILE * infile;
  struct jpeg_error_mgr jerr;
  struct jpeg_decompress_struct cinfo;
  JSAMPARRAY buffer_in = NULL;
  int row_stride = 0;
  char *buffer_out = NULL;

  if (argc != 2 ) {
    fprintf(stderr, "Usage: %s FILE\n", argv[0]);
    return -1;
  }
  file_out = argv[1];
    
  if ((infile = fopen("/dev/stdin", "r")) == NULL) {
    ERR("open %s: %s\n", file_in, strerror(errno));
    return -1;
  }

  cinfo.err = jpeg_std_error(&jerr);
  orig_output_message = cinfo.err->output_message;
  cinfo.err->output_message = libjpeg_output_message;
  jpeg_create_decompress(&cinfo);
  jpeg_stdio_src(&cinfo, infile);

  while(1) {

    jpeg_read_header(&cinfo, TRUE);
    jpeg_start_decompress(&cinfo);

    if (!row_stride) {
      row_stride = cinfo.output_width * cinfo.output_components;
      if (v4l(cinfo.output_width, cinfo.output_height) == -1)
	return -1;
      buffer_in = (*cinfo.mem->alloc_sarray)
	((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);
      buffer_out = malloc((cinfo.output_height + 1) * row_stride);
    }

    while(cinfo.output_scanline < cinfo.output_height) {
      (void) jpeg_read_scanlines(&cinfo, buffer_in, 1);
      memcpy(&buffer_out[cinfo.output_scanline * row_stride],
	     buffer_in[0], row_stride);
    }

    rgb(buffer_out, cinfo.output_height * row_stride);

    if (write(fd_out, buffer_out, cinfo.output_height * row_stride) !=
	cinfo.output_height * row_stride) {
      ERR("write error %s\n", strerror(errno));
      return -1;
    }

    jpeg_finish_decompress(&cinfo);
  }

  return 0;
}
