/* -*- linux-c -*-
 *
 * $Id: pccam880.c,v 1.15 2004/02/21 02:13:23 harmen Exp $ 
 *
 * Experimental driver for the Creative PC-CAM 880 USB webcam
 *
 * Copyright (C) 2004 Harmen van der Wal <harmen@harmwal.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/usb.h>                                                   
#include <linux/spinlock.h>
#include <asm/uaccess.h>


/* module parameters */
static int char_major = 60;
static int usb_vendor_id = 0x041e;
static int usb_device_id = 0x4024;
static int debug = 0;

#define WARN(fmt,args...) printk(KERN_WARNING "%s %s (%d) "fmt, __FILE__, __FUNCTION__, __LINE__,##args)
#define INFO(fmt,args...) printk(KERN_INFO "%s %s "fmt, __FILE__, __FUNCTION__,##args)
#define DBG(fmt,args...) if (debug) printk(KERN_DEBUG "(%ld) %s %s (%d) "fmt, jiffies, __FILE__, __FUNCTION__, __LINE__,##args)

                                                   
struct pccam880_camera {
	struct usb_device *udev;
	unsigned char bulk;
	int count;
	spinlock_t lock;
};

struct pccam880_camera *cam = NULL;

/* USB sniff */

//        in value   index   len         payload
static unsigned short init[][5] = {
	{ 0, 0x1f30, 0x0000,  0, 0 },
	{ 1, 0xd000, 0x0000,  4, 0 }, // 01 30 00 84
	{ 1, 0xf800, 0x0000, 10, 0 }, // 00 52 18 02 9b 01 0e 02 17 00
	{ 1, 0x9202, 0x0000,  2, 0 }, // 03 00
	{ 1, 0x9203, 0x0000, 12, 0 }, // 40 01 f0 00 a0 00 78 00  80 02 e0 01
	{ 0, 0x1f30, 0x0000,  0, 0 },
	{ 1, 0xd000, 0x0000,  4, 0 }, // 01 30 00 84
	{ 1, 0x0d00, 0x00aa,  2, 0 }, // 00 00
	{ 1, 0xbf01, 0x0000,  2, 0 }, // 6c 00
	{ 1, 0xbf00, 0x0000,  2, 0 }, // 00 00
	{ 1, 0xb601, 0x0000, 10, 0 }, // 00 00 00 02 20 00 b2 03 b0 03
	{ 0, 0xae00, 0x0000,  2, 1 }, // 01 00
	{ 1, 0xb600, 0x0000, 10, 0 }, // 00 00 00 00 1e 00 ec 00 00 00
	{}
};

static unsigned short open[][5] = {
	{ 0, 0x1f30, 0x0000,  0, 0  },
	{ 1, 0xd000, 0x0000,  4, 0  }, // 01 30 00 84
	{ 0, 0x3370, 0x0000, 10, 3  }, // 01 00 00 00 00 00 00 00 00 00
	{ 0, 0x2000, 0x0000,  0, 0  },
	{ 0, 0x2f0f, 0x0000,  0, 0  },
	{ 0, 0x2610, 0x0000,  6, 13 }, // 00 00 00 00 00 00
	{ 0, 0xe107, 0x0000,  0, 0  },
	{ 0, 0x2502, 0x0000,  0, 0  },
	{ 0, 0x1f70, 0x0000,  0, 0  },
	{ 1, 0xd000, 0x0000,  4, 0  }, // 01 70 00 84
	{ 0, 0x9a01, 0x0000,  2, 19 }, // 00 00
	{}
};

static unsigned short close[][5] = {
	{ 0, 0x1f30, 0x0000,  0, 0  },
	{ 1, 0xd000, 0x0000,  4, 0  }, // 01 30 00 00
	{}
};

static unsigned char payload[] = {
	0x00, // bogus

	// unknown
	0x01, 0x00,
	// image size
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	// compression rate
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	// unknown
	0x00, 0x00
};

static void pccam880_usb_msg(unsigned short data[][5])
{
	int dir, len, result, i = 0, ii;
	unsigned short value, index;
	unsigned char *buf, buf12[12];

	do {
		dir = data[i][0];
		value = data[i][1];
		index = data[i][2];
		len = data[i][3];

		buf = buf12;
		if (data[i][4] > 0)
			memcpy(buf,&payload[data[i][4]], len);
		if (len == 0)
			buf = NULL;

		result = usb_control_msg (
			cam->udev,
			dir ? 
			usb_rcvctrlpipe(cam->udev, 0) : 
			usb_sndctrlpipe(cam->udev, 0),
			1,
			(dir ?
			 USB_DIR_IN | USB_TYPE_VENDOR |USB_RECIP_DEVICE :
			 USB_DIR_OUT| USB_TYPE_VENDOR |USB_RECIP_DEVICE),
			value, index, buf, len, HZ);

		if (result < 0) {
		        WARN("result %d %s value 0x%04x index 0x%04x len %d",
			     result, dir ? "rcv" : "snd", value, index, len);
		} else {
		        DBG("result %d %s value 0x%04x index 0x%04x len %d",
			    result, dir ? "rcv" : "snd", value, index, len);
		}
		if (debug) {
		        for (ii = 0; ii < len; ii++)
				printk(" %02x", buf[ii]);
			printk("\n");
		}
	} while (value != 0x0000 && i++ < 16);
}

static int pccam880_open(struct inode *inode, struct file *filp)
{
	DBG("\n");

	spin_lock(&cam->lock);
	if (cam->count) {
		spin_unlock(&cam->lock);
		return -EBUSY;
	}
	cam->count++;
	spin_unlock(&cam->lock);

	pccam880_usb_msg(open);
	usb_clear_halt(cam->udev, usb_rcvbulkpipe(cam->udev, cam->bulk));

	return 0;
}

static ssize_t 
pccam880_read(struct file *filp, char *ubuf, size_t count, loff_t *f_pos)
{
	unsigned char buf[4096];
	int len, result;

	if (f_pos != &filp->f_pos)
		return -ESPIPE;

	if (ubuf == NULL)
		return -EINVAL;
	
	result = usb_bulk_msg(cam->udev,
			      usb_rcvbulkpipe(cam->udev, cam->bulk),
			      &buf, count, &len, HZ);

	if (result != 0)
		return result;

	if (copy_to_user(ubuf, buf, len))
		return -EFAULT;

	DBG("%d/%d\n", len, count);

	return len;
}

loff_t pccam880_llseek(struct file *filp, loff_t off, int whence)
{
	return -ESPIPE;
}

static int pccam880_release(struct inode *inode, struct file *filp)
{
	DBG("\n");

	pccam880_usb_msg(close);

	cam->count--;

	return 0;
}

static struct file_operations pccam880_fops = {
	open:  pccam880_open,
	read:  pccam880_read,
	llseek: pccam880_llseek,
	release: pccam880_release,
	owner: THIS_MODULE,
};

static void 
*pccam880_usb_probe(struct usb_device *dev, unsigned int ifnum, 
		    const struct usb_device_id *id)
{
	struct usb_interface_descriptor *interface;
	struct usb_endpoint_descriptor *endpoint;
	int result;

	if ((dev->descriptor.idVendor != usb_vendor_id) ||
	    (dev->descriptor.idProduct != usb_device_id)) {
		return NULL;
	}

	INFO("\n");

	if (cam) {
		WARN("multiple devices not supported\n");
		return NULL;
	}
	cam = kmalloc(sizeof (struct pccam880_camera), GFP_KERNEL);
	if (!cam) {
		WARN("can not kmalloc cam\n");
		return NULL;
	}
	memset(cam, 0, sizeof (struct pccam880_camera));
	cam->udev = dev;
	interface = &dev->actconfig->interface[ifnum].altsetting[0];
	endpoint = &interface->endpoint[0];
	cam->bulk = endpoint->bEndpointAddress;
	spin_lock_init(&cam->lock);

	result = register_chrdev(char_major, "pccam880", &pccam880_fops);
	if (result < 0) {
		WARN("can not register char_major %d\n", char_major);
		return NULL;
	}

	INFO("registered char_major %d\n", char_major);

	pccam880_usb_msg(init);

	return cam;
}

static void pccam880_usb_disconnect(struct usb_device *dev, void *ptr)
{
	INFO("\n");
	unregister_chrdev(char_major, "pccam880");
	kfree(cam);
	cam = NULL;
}

static struct usb_driver pccam880_usb_driver = {
	name: "pccam880",
	probe: pccam880_usb_probe,
	disconnect: pccam880_usb_disconnect,
};


MODULE_AUTHOR("Harmen van der Wal <harmen@harmwal.nl>");
MODULE_DESCRIPTION("Experimental driver for the Creative PC-CAM 880 USB webcam");
MODULE_LICENSE("GPL");
 
MODULE_PARM(char_major, "i");
MODULE_PARM(usb_vendor_id, "i");
MODULE_PARM(usb_device_id, "i");
MODULE_PARM(debug, "i");


static int pccam880_init(void)
{
	int result;

	INFO("\n");

	result = usb_register(&pccam880_usb_driver);
	if (result != 0) {
		WARN("usb_register failed\n");
		return result;
	}
	
	return 0;
}

static void pccam880_exit(void)
{
	INFO("\n");

	usb_deregister(&pccam880_usb_driver);
}

module_init(pccam880_init);
module_exit(pccam880_exit);


